/* Copyright 2000 Enhanced Software Technologies Inc.
 *   All Rights Reserved
 *
 * Released for public use under a BSD-style license. See file LICENSE
 * for details. 
 *
 *   Encrypt a file in CFB-128 mode using AES candidate "twofish".
 *
 * RCS CHANGE LOG:
 * $Log: aescrypt.c,v $
 * Revision 1.4  2000/04/24 19:46:57  eric
 * More autoconf tweaks. Changed to using Counterpane's TwoFish rather than
 * Dr. Gladman's TwoFish in an attempt to fix the big endian/little endian
 * problems once and for all.
 *
 * Revision 1.2  2000/04/05 22:09:19  eric
 * Daily checkin...
 *
 * Revision 1.1  2000/03/28 23:54:28  eric
 * Initial checkin -- aescrypt/aesget
 *
 * $Date: 2000/04/24 19:46:57 $
*/


#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "aes.h"
#include "bin2hex.h"
#include "dstring.h"
#include "dorandom.h"

static int infile;
static char *progname;

/* static unsigned char *key; /* the key to the program... */
keyInstance key;
cipherInstance cipher;

static int keysize;        /* and how big it is. */
static int cfb128_idx=-1;

static char *cfb_blk;
static char cfb_crypt[16];


static void usage(void) {
  fprintf(stderr,"%s:Usage: %s -k {keyfile}\n",progname,progname);
  exit(1);
}



/* write data, aborting if we have a write error. */

static void ewrite(int outfile,void *buf, size_t count) {
  int result;

  result=write(outfile,buf,count);
  if (result!=count) {
    fprintf(stderr,"%s:error:could not write data.\n",progname);
    perror(progname);
    usage();
  }
}

static void cryptblock(char *src, int srclen) {
  int i,ch;
 
  /* Now to encrypt each individual character in cfb mode: */
  for (i=0; i<srclen; i++) {
    if ((cfb128_idx < 0) || (cfb128_idx > 15)) {

      blockEncrypt(&cipher,&key,cfb_blk,128,cfb_crypt);
      /*encrypt(cfb_blk,cfb_crypt); */

      cfb128_idx=0;
    }
    /* XOR the data with a byte from our encrypted buffer. */ 
    ch=src[i] ^ cfb_crypt[cfb128_idx];
    /* do output feedback: put crypted byte into next block to be crypted */
    cfb_blk[cfb128_idx++]=ch; 
    src[i]=(char) ch;
  }
  
  return;
}


static void cryptfile(int infile,int outfile) {
  char *salt;
  int result;

#define CRYPTBUF_SIZE 8192  
  char buf[CRYPTBUF_SIZE];
  char outbuf[CRYPTBUF_SIZE];
  int bufsize;  /* how many chars are currently in that buffer. */

  result=urand_seed(NULL);
  if (result) {
    fprintf(stderr,"%s:error:Could not seed PRNG\n",progname);
    perror(progname);
    usage();
  }

  salt=urand_get(16);


  if (!salt) {
    fprintf(stderr,"%s:error:could not get salt from PRNG\n",progname);
    perror(progname);
    usage();
  }
  
  /* set the feedback buffer to be the salt. */
  cfb_blk=salt;

  /* we have salt... now write it to outfile.... */

  ewrite(outfile,salt,16);


  /* Now initialize the cipher: */
  if (cipherInit(&cipher,MODE_ECB,NULL) != TRUE) {
    fprintf(stderr,"%s:Could not initialize cipher.\n",progname);
    usage();
  }

  while (bufsize=read(infile,buf,CRYPTBUF_SIZE)) {
    if (bufsize<0) { /* some sort of I/O error... */
      fprintf(stderr,"%s:Error on read.\n",progname);
      perror(progname);
      exit(1);
    }

    cryptblock(buf,bufsize);
    ewrite(outfile,buf,bufsize); /* checks for error... */
  }

  /* okay, we've done everything... */
  
  return;
}
  


/* Routine to set the key once we've read it. */
static void aes_set_key(char *inkey) {
  int keylen,i;
  char *hexkey;
 char *binkey; 

  hexkey=d_trim(inkey); 

  keylen=strlen(hexkey);

  if ( keylen==32||keylen==64||keylen==48) {
    keysize=keylen>>1;
    
    binkey=unhexify_block(hexkey,keysize);
    if (!binkey) {
      fprintf(stderr,"%s:Key format error:%s\n",progname,hexkey);
      usage();
    }

    /* now to actually set the key... */
    if (makeKey(&key,DIR_ENCRYPT,keylen*8,NULL)!=TRUE) {
      fprintf(stderr,"%s:Key format error:%s\n",progname,hexkey);
      usage();
    }
    /* Now copy the actual binary key, sigh. */
    for (i=0;i<keysize;i++) {
      key.key32[i]=binkey[i];
    }
    reKey(&key);   /* run the key schedule. */
      
    return;
  }
    
  fprintf(stderr,"%s:Key size error:%d[%s]\n",progname,keylen,hexkey);
  usage();
}
  
/* Routine to read a key from keyfile. */
static void read_key(char *filename) {
  char inbuf[1024];
  FILE *infile;
  char *result;

  infile=fopen(filename,"r");
  if (!infile) {
    fprintf(stderr,"%s:Error: Key file '%s' not found.\n",progname,filename);
    usage();
  }

  while (result=fgets(inbuf,1023,infile)) {
    if (strncmp(result,"kk:",3)==0) {
      aes_set_key(result+3);
      return;
    }
  }

  fprintf(stderr,"%s: Error: Key not found in file '%s'.\n",progname,filename);
  usage();
}
  


int main(int argc, char **argv) {

  progname= argv[0];
  if (argc < 2) {
    usage();
  }

  if (strcmp(argv[1],"-k")) {
    usage();
  }

  read_key(argv[2]); /* read the keyfile, hopefully! */

  cryptfile(fileno(stdin),fileno(stdout));
  
  
  exit(0);

}
